#include "Kconst.hpp"
#include "KLNode.hpp"
#include "KINode.hpp"

KINode::KINode(pKINode parent, pRect rect) : KNode(false, parent, rect)
{
	//-- reserve space for children --
	this->child = new KLNode[KNODE_SZ2];

	//-- init children --
	double x_sz = rect->x_sz / KNODE_SZ;
	double y_sz = rect->y_sz / KNODE_SZ;

	for (int y = 0; y < KNODE_SZ; y++)
	{
		pKLNode lnode = &child[y * KNODE_SZ];

		for (int x = 0; x < KNODE_SZ; x++)
		{
			//-- KNode part --
			lnode->isLeaf = true;
			lnode->parent = this;
			lnode->rect.x = rect->x + x * x_sz;
			lnode->rect.y = rect->y + y * y_sz;
			lnode->rect.x_sz = x_sz;
			lnode->rect.y_sz = y_sz;
			lnode->fullQ = new DynArray<pQuery>();

			//-- KLNode part --
			lnode->partQ = new DynArray<pQuery>();


			//-- next leaf --
			lnode ++;
		}
	}

}

void KINode::trunc(int &x)
{
	if (x < 0) x = 0;
	if (x >= KNODE_SZ) x = KNODE_SZ - 1;
}

void KINode::add(pQuery q)
{
	//-- get coordinates of q relative to rect --
	double x1 = q->x - rect.x;
	double y1 = q->y - rect.y;
	double x2 = q->x + q->x_sz - rect.x;
	double y2 = q->y + q->y_sz - rect.y;

	//-- compute q's cell boundaries --
	double x_sz = rect.x_sz / KNODE_SZ;
	double y_sz = rect.y_sz / KNODE_SZ;

	int cell_x1 = (int) (x1 / x_sz);
	int cell_y1 = (int) (y1 / y_sz);
	int cell_x2 = (int) (x2 / x_sz);
	int cell_y2 = (int) (y2 / y_sz);

	trunc(cell_x1);
	trunc(cell_y1);
	trunc(cell_x2);
	trunc(cell_y2);

	//-- add queries to cells --
	for (int y = cell_y1; y <= cell_y2; y++)
	{
		pKLNode lnode = &child[y * KNODE_SZ + cell_x1];

		for (int x = cell_x1; x <= cell_x2; x++)
		{
			pKNode node = (pKNode) lnode;
			node->add(q);
			lnode ++;
		}
	}
}
