#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "../data/Rect.hpp"
#include "../struc/ListNode.hpp"
#include "../struc/List.hpp"

#include "Kconst.hpp"
#include "KNode.hpp"
#include "KLNode.hpp"
#include "KINode.hpp"

#include "KTree.hpp"


//--------------------------------------------
KTree::KTree()
{
	Rect rect;

	rect.x    = 0;
	rect.y    = 0;
	rect.x_sz = 1;
	rect.y_sz = 1;

	root = new KINode(NULL, &rect);


	if (sizeof (KLNode) != sizeof (KINode))
	{
		printf("\nKTree::KTree(): q1");
		exit(-1);
	}
}

//--------------------------------------------
DynArray<pQuery>* KTree::process(pPoint p)
{
	static DynArray<pQuery> *qarr = new DynArray<pQuery>(1000);
	qarr->cur_sz = 0;
	
	pKNode node = root;

	while(true)
	{
		pKINode inode = (pKINode)node;

		for (int i = 0; i < inode->fullQ->cur_sz; i++)
		{
			pQuery q = inode->fullQ->arr[i];
			qarr->insert(q);
		}

		//-- get p's relative coordinates --
		pRect rect = &node->rect;

		double x = p->x - rect->x;
		double y = p->y - rect->y;
		double x_sz = rect->x_sz / KNODE_SZ;
		double y_sz = rect->y_sz / KNODE_SZ;

		//-- determine p's cell --
		int cell_x = (int) (x / x_sz);
		int cell_y = (int) (y / y_sz);

		//if (cell_x < 0) cell_x = 0;
		//if (cell_x >= KNODE_SZ) cell_x = KNODE_SZ - 1;
		//if (cell_y < 0) cell_y = 0;
		//if (cell_y >= KNODE_SZ) cell_y = KNODE_SZ - 1;

		if (node->isLeaf)
			break;

		node = &inode->child[cell_y * KNODE_SZ + cell_x];
	}

	//-- leaf node --
	pKLNode lnode = (pKLNode) node;

	for (int i = 0; i < lnode->partQ->cur_sz; i++)
	{
		pQuery q = lnode->partQ->arr[i];

		if (q->intersectsWith(p))
			qarr->insert(q);
	}


	return qarr;
}

//--------------------------------------------
void KTree::getQ2P(Point *pointArr, int num_point, Query *queryArr, int num_query, pPListNode q2pmap)
{
	int i;

	//-- kill previous mapping --
	int map_len =0;

	Query *q = &queryArr[0]; 

	for (i = 0; i < num_query; i++)
	{
		q->q2p_list = NULL;
		q ++;
	}
	
	//-- create new mapping: do point by point processing --
	Point *p = &pointArr[0]; 
	
	for (int k = 0; k < num_point; k++) 
	{ 
		//-- get KTree's leaf node for that point -- 
		DynArray<pQuery> *qarr = process(p);
		
		for (int i = 0; i < qarr->cur_sz; i++)
		{ 
			pQuery q = qarr->arr[i];

			//-- if  q contains point p => add p to q'a answer list -- 
			if ((q->x <= p->x) &&
				(q->y <= p->y) &&
				(q->x + q->x_sz >= p->x) &&
				(q->y + q->y_sz >= p->y) )
			{
				//-- create new element in q2p map --
				pPListNode map_el = &q2pmap[map_len];
				map_len ++;
				
				//-- adjust new map element --
				map_el->data = p;
				map_el->next = q->q2p_list;
				
				//-- adjust q's q2p list
				q->q2p_list = map_el; 
				
			}
		} 
		
	
		//-- next point --
		p ++;
	}
	
	printf("\n Q->P only (map_len = %d)", map_len);
}


//--------------------------------------------
void KTree::addQueryArr(pQuery queryArr, int num_query)
{
	for (int i = 0; queryArr[i].x != -1; i++)
	{
		//printf("\nadding q#%d", i);

		//if (i == 9)
			//printf("i=%d", i);

		add(&queryArr[i]);
	}
}


//--------------------------------------------
void KTree::add(pQuery q)
{
	root->add(q);
}

//--------------------------------------------
void KTree::remove(pQuery q)
{
	printf("\nKTree::remove() not implemented");
	getchar();
}


